Developer

This document explains how you can develop modules for the Apache HTTP
Server 2.4

Request Processing in Apache 2.4

Apache 2.x Hook Functions

Introduction
============

What we will be discussing in this document {#what}
-------------------------------------------

This document will discuss how you can create modules for the Apache
HTTP Server 2.4, by exploring an example module called

mod\_example

. In the first part of this document, the purpose of this module will be
to calculate and print out various digest values for existing files on
your web server, whenever we access the URL

http://hostname/filename.sum

. For instance, if we want to know the MD5 digest value of the file
located at

http://www.example.com/index.html

, we would visit

http://www.example.com/index.html.sum

.

In the second part of this document, which deals with configuration
directive and context awareness, we will be looking at a module that
simply writes out its own configuration to the client.

Prerequisites
-------------

First and foremost, you are expected to have a basic knowledge of how
the C programming language works. In most cases, we will try to be as
pedagogical as possible and link to documents describing the functions
used in the examples, but there are also many cases where it is
necessary to either just assume that \"it works\" or do some digging
yourself into what the hows and whys of various function calls.

Lastly, you will need to have a basic understanding of how modules are
loaded and configured in the Apache HTTP Server, as well as how to get
the headers for Apache if you do not have them already, as these are
needed for compiling new modules.

Compiling your module {#compiling}
---------------------

To compile the source code we are building in this document, we will be
using

APXS

. Assuming your source file is called mod\_example.c, compiling,
installing and activating the module is as simple as:

apxs -i -a -c mod\_example.c

Defining a module {#basics}
=================

Every module starts with the same declaration, or name tag if you will,
that defines a module as

a separate entity within Apache

:

module AP\_MODULE\_DECLARE\_DATA example\_module = {
STANDARD20\_MODULE\_STUFF, create\_dir\_conf, /\* Per-directory
configuration handler \*/ merge\_dir\_conf, /\* Merge handler for
per-directory configurations \*/ create\_svr\_conf, /\* Per-server
configuration handler \*/ merge\_svr\_conf, /\* Merge handler for
per-server configurations \*/ directives, /\* Any directives we may have
for httpd \*/ register\_hooks /\* Our hook registering function \*/ };

This bit of code lets the server know that we have now registered a new
module in the system, and that its name is

example\_module

. The name of the module is used primarily for two things:

Letting the server know how to load the module using the LoadModule

Setting up a namespace for the module to use in configurations

For now, we\'re only concerned with the first purpose of the module
name, which comes into play when we need to load the module:

LoadModule example\_module \"modules/mod\_example.so\"

In essence, this tells the server to open up

mod\_example.so

and look for a module called

example\_module

.

Within this name tag of ours is also a bunch of references to how we
would like to handle things: Which directives do we respond to in a
configuration file or .htaccess, how do we operate within specific
contexts, and what handlers are we interested in registering with the
Apache HTTP service. We\'ll return to all these elements later in this
document.

Getting started: Hooking into the server {#hooking}
========================================

An introduction to hooks {#hook_intro}
------------------------

When handling requests in Apache HTTP Server 2.4, the first thing you
will need to do is create a hook into the request handling process. A
hook is essentially a message telling the server that you are willing to
either serve or at least take a glance at certain requests given by
clients. All handlers, whether it\'s mod\_rewrite, mod\_authn\_\*,
mod\_proxy and so on, are hooked into specific parts of the request
process. As you are probably aware, modules serve different purposes;
Some are authentication/authorization handlers, others are file or
script handlers while some third modules rewrite URIs or proxies
content. Furthermore, in the end, it is up to the user of the server how
and when each module will come into place. Thus, the server itself does
not presume to know which module is responsible for handling a specific
request, and will ask each module whether they have an interest in a
given request or not. It is then up to each module to either gently
decline serving a request, accept serving it or flat out deny the
request from being served, as authentication/authorization modules do:

To make it a bit easier for handlers such as our mod\_example to know
whether the client is requesting content we should handle or not, the
server has directives for hinting to modules whether their assistance is
needed or not. Two of these are

AddHandler

and

SetHandler

. Let\'s take a look at an example using

AddHandler

. In our example case, we want every request ending with .sum to be
served by

mod\_example

, so we\'ll add a configuration directive that tells the server to do
just that:

AddHandler example-handler \".sum\"

What this tells the server is the following:

Whenever we receive a request for a URI ending in .sum, we are to let
all modules know that we are looking for whoever goes by the name of
\"example-handler\"

. Thus, when a request is being served that ends in .sum, the server
will let all modules know, that this request should be served by
\"example-handler\". As you will see later, when we start building
mod\_example, we will check for this handler tag relayed by

AddHandler

and reply to the server based on the value of this tag.

Hooking into httpd {#hook_declaration}
------------------

To begin with, we only want to create a simple handler that replies to
the client browser when a specific URL is requested, so we won\'t bother
setting up configuration handlers and directives just yet. Our initial
module definition will look like this:

module AP\_MODULE\_DECLARE\_DATA example\_module = {
STANDARD20\_MODULE\_STUFF, NULL, NULL, NULL, NULL, NULL, register\_hooks
/\* Our hook registering function \*/ };

This lets the server know that we are not interested in anything fancy,
we just want to hook onto the requests and possibly handle some of them.

The reference in our example declaration,

register\_hooks

is the name of a function we will create to manage how we hook onto the
request process. In this example module, the function has just one
purpose; To create a simple hook that gets called after all the
rewrites, access control etc has been handled. Thus, we will let the
server know that we want to hook into its process as one of the last
modules:

static void register\_hooks(apr\_pool\_t \*pool) { /\* Create a hook in
the request handler, so we get called when a request arrives \*/
ap\_hook\_handler(example\_handler, NULL, NULL, APR\_HOOK\_LAST); }

The

example\_handler

reference is the function that will handle the request. We will discuss
how to create a handler in the next chapter.

Other useful hooks {#hook_others}
------------------

Hooking into the request handling phase is but one of many hooks that
you can create. Some other ways of hooking are:

ap\_hook\_child\_init

: Place a hook that executes when a child process is spawned (commonly
used for initializing modules after the server has forked)

ap\_hook\_pre\_config

: Place a hook that executes before any configuration data has been read
(very early hook)

ap\_hook\_post\_config

: Place a hook that executes after configuration has been parsed, but
before the server has forked

ap\_hook\_pre\_translate\_name

: Place a hook that executes when a URI needs to be translated into a
filename on the server, before decoding

ap\_hook\_translate\_name

: Place a hook that executes when a URI needs to be translated into a
filename on the server (think

mod\_rewrite

)

ap\_hook\_quick\_handler

: Similar to

ap\_hook\_handler

, except it is run before any other request hooks (translation, auth,
fixups etc)

ap\_hook\_log\_transaction

: Place a hook that executes when the server is about to add a log entry
of the current request

Building a handler {#handling}
==================

A handler is essentially a function that receives a callback when a
request to the server is made. It is passed a record of the current
request (how it was made, which headers and requests were passed along,
who\'s giving the request and so on), and is put in charge of either
telling the server that it\'s not interested in the request or handle
the request with the tools provided.

A simple \"Hello, world!\" handler {#simple_handler}
----------------------------------

Let\'s start off by making a very simple request handler that does the
following:

Check that this is a request that should be served by
\"example-handler\"

Set the content type of our output to

text/html

Write \"Hello, world!\" back to the client browser

Let the server know that we took care of this request and everything
went fine

In C code, our example handler will now look like this:

static int example\_handler(request\_rec \*r) { /\* First off, we need
to check if this is a call for the \"example-handler\" handler. \* If it
is, we accept it and do our things, if not, we simply return DECLINED,
\* and the server will try somewhere else. \*/ if (!r-\>handler \|\|
strcmp(r-\>handler, \"example-handler\")) return (DECLINED); /\* Now
that we are handling this request, we\'ll write out \"Hello, world!\" to
the client. \* To do so, we must first set the appropriate content type,
followed by our output. \*/ ap\_set\_content\_type(r, \"text/html\");
ap\_rprintf(r, \"Hello, world!\"); /\* Lastly, we must tell the server
that we took care of this request and everything went fine. \* We do so
by simply returning the value OK to the server. \*/ return OK; }

Now, we put all we have learned together and end up with a program that
looks like

mod\_example\_1.c

. The functions used in this example will be explained later in the
section

\"Some useful functions you should know\"

.

The request\_rec structure {#request_rec}
--------------------------

The most essential part of any request is the

request record

. In a call to a handler function, this is represented by the

request\_rec\*

structure passed along with every call that is made. This struct,
typically just referred to as

r

in modules, contains all the information you need for your module to
fully process any HTTP request and respond accordingly.

Some key elements of the

request\_rec

structure are:

r-\>handler (char\*):

Contains the name of the handler the server is currently asking to do
the handling of this request

r-\>method (char\*):

Contains the HTTP method being used, f.x. GET or POST

r-\>filename (char\*):

Contains the translated filename the client is requesting

r-\>args (char\*):

Contains the query string of the request, if any

r-\>headers\_in (apr\_table\_t\*):

Contains all the headers sent by the client

r-\>connection (conn\_rec\*):

A record containing information about the current connection

r-\>user (char\*):

If the URI requires authentication, this is set to the username provided

r-\>useragent\_ip (char\*):

The IP address of the client connecting to us

r-\>pool (apr\_pool\_t\*)

: The memory pool of this request. We\'ll discuss this in the \"

Memory management

\" chapter.

A complete list of all the values contained within the

request\_rec

structure can be found in the

httpd.h

header file or at

http://ci.apache.org/projects/httpd/trunk/doxygen/structrequest\_\_rec.html

.

Let\'s try out some of these variables in another example handler:

static int example\_handler(request\_rec \*r) { /\* Set the appropriate
content type \*/ ap\_set\_content\_type(r, \"text/html\"); /\* Print out
the IP address of the client connecting to us: \*/ ap\_rprintf(r,
\"\<h2\>Hello, %s!\</h2\>\", r-\>useragent\_ip); /\* If we were reached
through a GET or a POST request, be happy, else sad. \*/ if (
!strcmp(r-\>method, \"POST\") \|\| !strcmp(r-\>method, \"GET\") ) {
ap\_rputs(\"You used a GET or a POST method, that makes us
happy!\<br/\>\", r); } else { ap\_rputs(\"You did not use POST or GET,
that makes us sad :(\<br/\>\", r); } /\* Lastly, if there was a query
string, let\'s print that too! \*/ if (r-\>args) { ap\_rprintf(r, \"Your
query string was: %s\", r-\>args); } return OK; }

Return values {#return_value}
-------------

Apache relies on return values from handlers to signify whether a
request was handled or not, and if so, whether the request went well or
not. If a module is not interested in handling a specific request, it
should always return the value

DECLINED

. If it is handling a request, it should either return the generic value

OK

, or a specific HTTP status code, for example:

static int example\_handler(request\_rec \*r) { /\* Return 404: Not
found \*/ return HTTP\_NOT\_FOUND; }

Returning

OK

or a HTTP status code does not necessarily mean that the request will
end. The server may still have other handlers that are interested in
this request, for instance the logging modules which, upon a successful
request, will write down a summary of what was requested and how it
went. To do a full stop and prevent any further processing after your
module is done, you can return the value

DONE

to let the server know that it should cease all activity on this request
and carry on with the next, without informing other handlers.

General response codes:

DECLINED

: We are not handling this request

OK

: We handled this request and it went well

DONE

: We handled this request and the server should just close this thread
without further processing

HTTP specific return codes (excerpt):

HTTP\_OK (200)

: Request was okay

HTTP\_MOVED\_PERMANENTLY (301)

: The resource has moved to a new URL

HTTP\_UNAUTHORIZED (401)

: Client is not authorized to visit this page

HTTP\_FORBIDDEN (403)

: Permission denied

HTTP\_NOT\_FOUND (404)

: File not found

HTTP\_INTERNAL\_SERVER\_ERROR (500)

: Internal server error (self explanatory)

Some useful functions you should know {#functions}
-------------------------------------

ap\_rputs(const char \*string, request\_rec \*r)

:

Sends a string of text to the client. This is a shorthand version of

ap\_rwrite

.

ap\_rputs(\"Hello, world!\", r);

ap\_rprintf

:

This function works just like

printf

, except it sends the result to the client.

ap\_rprintf(r, \"Hello, %s!\", r-\>useragent\_ip);

ap\_set\_content\_type

(request\_rec \*r, const char \*type)

:

Sets the content type of the output you are sending.

ap\_set\_content\_type(r, \"text/plain\"); /\* force a raw text output
\*/

Memory management {#memory}
-----------------

Managing your resources in Apache HTTP Server 2.4 is quite easy, thanks
to the memory pool system. In essence, each server, connection and
request have their own memory pool that gets cleaned up when its scope
ends, e.g. when a request is done or when a server process shuts down.
All your module needs to do is latch onto this memory pool, and you
won\'t have to worry about having to clean up after yourself - pretty
neat, huh?

In our module, we will primarily be allocating memory for each request,
so it\'s appropriate to use the

r-\>pool

reference when creating new objects. A few of the functions for
allocating memory within a pool are:

void\*

apr\_palloc

( apr\_pool\_t \*p, apr\_size\_t size)

: Allocates

size

number of bytes in the pool for you

void\*

apr\_pcalloc

( apr\_pool\_t \*p, apr\_size\_t size)

: Allocates

size

number of bytes in the pool for you and sets all bytes to 0

char\*

apr\_pstrdup

( apr\_pool\_t \*p, const char \*s)

: Creates a duplicate of the string

s

. This is useful for copying constant values so you can edit them

char\*

apr\_psprintf

( apr\_pool\_t \*p, const char \*fmt, \...)

: Similar to

sprintf

, except the server supplies you with an appropriately allocated target
variable

Let\'s put these functions into an example handler:

static int example\_handler(request\_rec \*r) { const char \*original =
\"You can\'t edit this!\"; char \*copy; int \*integers; /\* Allocate
space for 10 integer values and set them all to zero. \*/ integers =
apr\_pcalloc(r-\>pool, sizeof(int)\*10); /\* Create a copy of the
\'original\' variable that we can edit. \*/ copy =
apr\_pstrdup(r-\>pool, original); return OK; }

This is all well and good for our module, which won\'t need any
pre-initialized variables or structures. However, if we wanted to
initialize something early on, before the requests come rolling in, we
could simply add a call to a function in our

register\_hooks

function to sort it out:

static void register\_hooks(apr\_pool\_t \*pool) { /\* Call a function
that initializes some stuff \*/ example\_init\_function(pool); /\*
Create a hook in the request handler, so we get called when a request
arrives \*/ ap\_hook\_handler(example\_handler, NULL, NULL,
APR\_HOOK\_LAST); }

In this pre-request initialization function we would not be using the
same pool as we did when allocating resources for request-based
functions. Instead, we would use the pool given to us by the server for
allocating memory on a per-process based level.

Parsing request data {#parsing}
--------------------

In our example module, we would like to add a feature, that checks which
type of digest, MD5 or SHA1 the client would like to see. This could be
solved by adding a query string to the request. A query string is
typically comprised of several keys and values put together in a string,
for instance

valueA=yes&valueB=no&valueC=maybe

. It is up to the module itself to parse these and get the data it
requires. In our example, we\'ll be looking for a key called

digest

, and if set to

md5

, we\'ll produce an MD5 digest, otherwise we\'ll produce a SHA1 digest.

Since the introduction of Apache HTTP Server 2.4, parsing request data
from GET and POST requests have never been easier. All we require to
parse both GET and POST data is four simple lines:

apr\_table\_t

\*GET;

apr\_array\_header\_t

\*POST;

ap\_args\_to\_table

(r, &GET);

ap\_parse\_form\_data

(r, NULL, &POST, -1, 8192);

In our specific example module, we\'re looking for the

digest

value from the query string, which now resides inside a table called

GET

. To extract this value, we need only perform a simple operation:

/\* Get the \"digest\" key from the query string, if any. \*/ const char
\*digestType = apr\_table\_get(GET, \"digest\"); /\* If no key was
returned, we will set a default value instead. \*/ if (!digestType)
digestType = \"sha1\";

The structures used for the POST and GET data are not exactly the same,
so if we were to fetch a value from POST data instead of the query
string, we would have to resort to a few more lines, as outlined in

this example

in the last chapter of this document.

Making an advanced handler {#advanced_handler}
--------------------------

Now that we have learned how to parse form data and manage our
resources, we can move on to creating an advanced version of our module,
that spits out the MD5 or SHA1 digest of files:

static int example\_handler(request\_rec \*r) { int rc, exists;
apr\_finfo\_t finfo; apr\_file\_t \*file; char \*filename; char
buffer\[256\]; apr\_size\_t readBytes; int n; apr\_table\_t \*GET;
apr\_array\_header\_t \*POST; const char \*digestType; /\* Check that
the \"example-handler\" handler is being called. \*/ if (!r-\>handler
\|\| strcmp(r-\>handler, \"example-handler\")) return (DECLINED); /\*
Figure out which file is being requested by removing the .sum from it
\*/ filename = apr\_pstrdup(r-\>pool, r-\>filename);
filename\[strlen(filename)-4\] = 0; /\* Cut off the last 4 characters.
\*/ /\* Figure out if the file we request a sum on exists and isn\'t a
directory \*/ rc = apr\_stat(&finfo, filename, APR\_FINFO\_MIN,
r-\>pool); if (rc == APR\_SUCCESS) { exists = ( (finfo.filetype !=
APR\_NOFILE) && !(finfo.filetype & APR\_DIR) ); if (!exists) return
HTTP\_NOT\_FOUND; /\* Return a 404 if not found. \*/ } /\* If apr\_stat
failed, we\'re probably not allowed to check this file. \*/ else return
HTTP\_FORBIDDEN; /\* Parse the GET and, optionally, the POST data sent
to us \*/ ap\_args\_to\_table(r, &GET); ap\_parse\_form\_data(r, NULL,
&POST, -1, 8192); /\* Set the appropriate content type \*/
ap\_set\_content\_type(r, \"text/html\"); /\* Print a title and some
general information \*/ ap\_rprintf(r, \"\<h2\>Information on
%s:\</h2\>\", filename); ap\_rprintf(r, \"\<b\>Size:\</b\> %u
bytes\<br/\>\", finfo.size); /\* Get the digest type the client wants to
see \*/ digestType = apr\_table\_get(GET, \"digest\"); if (!digestType)
digestType = \"MD5\"; rc = apr\_file\_open(&file, filename, APR\_READ,
APR\_OS\_DEFAULT, r-\>pool); if (rc == APR\_SUCCESS) { /\* Are we trying
to calculate the MD5 or the SHA1 digest? \*/ if (!strcasecmp(digestType,
\"md5\")) { /\* Calculate the MD5 sum of the file \*/ union { char
chr\[16\]; uint32\_t num\[4\]; } digest; apr\_md5\_ctx\_t md5;
apr\_md5\_init(&md5); readBytes = 256; while ( apr\_file\_read(file,
buffer, &readBytes) == APR\_SUCCESS ) { apr\_md5\_update(&md5, buffer,
readBytes); } apr\_md5\_final(digest.chr, &md5); /\* Print out the MD5
digest \*/ ap\_rputs(\"\<b\>MD5: \</b\>\<code\>\", r); for (n = 0; n \<
APR\_MD5\_DIGESTSIZE/4; n++) { ap\_rprintf(r, \"%08x\",
digest.num\[n\]); } ap\_rputs(\"\</code\>\", r); /\* Print a link to the
SHA1 version \*/ ap\_rputs(\"\<br/\>\<a href=\'?digest=sha1\'\>View the
SHA1 hash instead\</a\>\", r); } else { /\* Calculate the SHA1 sum of
the file \*/ union { char chr\[20\]; uint32\_t num\[5\]; } digest;
apr\_sha1\_ctx\_t sha1; apr\_sha1\_init(&sha1); readBytes = 256; while (
apr\_file\_read(file, buffer, &readBytes) == APR\_SUCCESS ) {
apr\_sha1\_update(&sha1, buffer, readBytes); }
apr\_sha1\_final(digest.chr, &sha1); /\* Print out the SHA1 digest \*/
ap\_rputs(\"\<b\>SHA1: \</b\>\<code\>\", r); for (n = 0; n \<
APR\_SHA1\_DIGESTSIZE/4; n++) { ap\_rprintf(r, \"%08x\",
digest.num\[n\]); } ap\_rputs(\"\</code\>\", r); /\* Print a link to the
MD5 version \*/ ap\_rputs(\"\<br/\>\<a href=\'?digest=md5\'\>View the
MD5 hash instead\</a\>\", r); } apr\_file\_close(file); } /\* Let the
server know that we responded to this request. \*/ return OK; }

This version in its entirety can be found here:

mod\_example\_2.c

.

Adding configuration options {#configuration}
============================

In this next segment of this document, we will turn our eyes away from
the digest module and create a new example module, whose only function
is to write out its own configuration. The purpose of this is to examine
how the server works with configuration, and what happens when you start
writing advanced configurations for your modules.

An introduction to configuration directives {#config_intro}
-------------------------------------------

If you are reading this, then you probably already know what a
configuration directive is. Simply put, a directive is a way of telling
an individual module (or a set of modules) how to behave, such as these
directives control how

mod\_rewrite

works:

RewriteEngine On RewriteCond \"%{REQUEST\_URI}\" \"\^/foo/bar\"
RewriteRule \"\^/foo/bar/(.\*)\$\" \"/foobar?page=\$1\"

Each of these configuration directives are handled by a separate
function, that parses the parameters given and sets up a configuration
accordingly.

Making an example configuration {#config_simple}
-------------------------------

To begin with, we\'ll create a basic configuration in C-space:

typedef struct { int enabled; /\* Enable or disable our module \*/ const
char \*path; /\* Some path to\...something \*/ int typeOfAction; /\* 1
means action A, 2 means action B and so on \*/ } example\_config;

Now, let\'s put this into perspective by creating a very small module
that just prints out a hard-coded configuration. You\'ll notice that we
use the

register\_hooks

function for initializing the configuration values to their defaults:

typedef struct { int enabled; /\* Enable or disable our module \*/ const
char \*path; /\* Some path to\...something \*/ int typeOfAction; /\* 1
means action A, 2 means action B and so on \*/ } example\_config; static
example\_config config; static int example\_handler(request\_rec \*r) {
if (!r-\>handler \|\| strcmp(r-\>handler, \"example-handler\"))
return(DECLINED); ap\_set\_content\_type(r, \"text/plain\");
ap\_rprintf(r, \"Enabled: %u\\n\", config.enabled); ap\_rprintf(r,
\"Path: %s\\n\", config.path); ap\_rprintf(r, \"TypeOfAction: %x\\n\",
config.typeOfAction); return OK; } static void
register\_hooks(apr\_pool\_t \*pool) { config.enabled = 1; config.path =
\"/foo/bar\"; config.typeOfAction = 0x00;
ap\_hook\_handler(example\_handler, NULL, NULL, APR\_HOOK\_LAST); } /\*
Define our module as an entity and assign a function for registering
hooks \*/ module AP\_MODULE\_DECLARE\_DATA example\_module = {
STANDARD20\_MODULE\_STUFF, NULL, /\* Per-directory configuration handler
\*/ NULL, /\* Merge handler for per-directory configurations \*/ NULL,
/\* Per-server configuration handler \*/ NULL, /\* Merge handler for
per-server configurations \*/ NULL, /\* Any directives we may have for
httpd \*/ register\_hooks /\* Our hook registering function \*/ };

So far so good. To access our new handler, we could add the following to
our configuration:

\<Location \"/example\"\> SetHandler example-handler \</Location\>

When we visit, we\'ll see our current configuration being spit out by
our module.

Registering directives with the server {#register_directive}
--------------------------------------

What if we want to change our configuration, not by hard-coding new
values into the module, but by using either the httpd.conf file or
possibly a .htaccess file? It\'s time to let the server know that we
want this to be possible. To do so, we must first change our

name tag

to include a reference to the configuration directives we want to
register with the server:

module AP\_MODULE\_DECLARE\_DATA example\_module = {
STANDARD20\_MODULE\_STUFF, NULL, /\* Per-directory configuration handler
\*/ NULL, /\* Merge handler for per-directory configurations \*/ NULL,
/\* Per-server configuration handler \*/ NULL, /\* Merge handler for
per-server configurations \*/ example\_directives, /\* Any directives we
may have for httpd \*/ register\_hooks /\* Our hook registering function
\*/ };

This will tell the server that we are now accepting directives from the
configuration files, and that the structure called

example\_directives

holds information on what our directives are and how they work. Since we
have three different variables in our module configuration, we will add
a structure with three directives and a NULL at the end:

static const command\_rec example\_directives\[\] = {
AP\_INIT\_TAKE1(\"exampleEnabled\", example\_set\_enabled, NULL,
RSRC\_CONF, \"Enable or disable mod\_example\"),
AP\_INIT\_TAKE1(\"examplePath\", example\_set\_path, NULL, RSRC\_CONF,
\"The path to whatever\"), AP\_INIT\_TAKE2(\"exampleAction\",
example\_set\_action, NULL, RSRC\_CONF, \"Special action value!\"), {
NULL } };

As you can see, each directive needs at least 5 parameters set:

AP\_INIT\_TAKE1

: This is a macro that tells the server that this directive takes one
and only one argument. If we required two arguments, we could use the
macro

AP\_INIT\_TAKE2

and so on (refer to httpd\_conf.h for more macros).

exampleEnabled

: This is the name of our directive. More precisely, it is what the user
must put in his/her configuration in order to invoke a configuration
change in our module.

example\_set\_enabled

: This is a reference to a C function that parses the directive and sets
the configuration accordingly. We will discuss how to make this in the
following paragraph.

RSRC\_CONF

: This tells the server where the directive is permitted. We\'ll go into
details on this value in the later chapters, but for now,

RSRC\_CONF

means that the server will only accept these directives in a server
context.

\"Enable or disable\....\"

: This is simply a brief description of what the directive does.

(

The \"missing\" parameter in our definition, which is usually set to

NULL

, is an optional function that can be run after the initial function to
parse the arguments have been run. This is usually omitted, as the
function for verifying arguments might as well be used to set them.

)

The directive handler function {#directive_handler}
------------------------------

Now that we have told the server to expect some directives for our
module, it\'s time to make a few functions for handling these. What the
server reads in the configuration file(s) is text, and so naturally,
what it passes along to our directive handler is one or more strings,
that we ourselves need to recognize and act upon. You\'ll notice, that
since we set our

exampleAction

directive to accept two arguments, its C function also has an additional
parameter defined:

/\* Handler for the \"exampleEnabled\" directive \*/ const char
\*example\_set\_enabled(cmd\_parms \*cmd, void \*cfg, const char \*arg)
{ if(!strcasecmp(arg, \"on\")) config.enabled = 1; else config.enabled =
0; return NULL; } /\* Handler for the \"examplePath\" directive \*/
const char \*example\_set\_path(cmd\_parms \*cmd, void \*cfg, const char
\*arg) { config.path = arg; return NULL; } /\* Handler for the
\"exampleAction\" directive \*/ /\* Let\'s pretend this one takes one
argument (file or db), and a second (deny or allow), \*/ /\* and we
store it in a bit-wise manner. \*/ const char
\*example\_set\_action(cmd\_parms \*cmd, void \*cfg, const char \*arg1,
const char \*arg2) { if(!strcasecmp(arg1, \"file\")) config.typeOfAction
= 0x01; else config.typeOfAction = 0x02; if(!strcasecmp(arg2, \"deny\"))
config.typeOfAction += 0x10; else config.typeOfAction += 0x20; return
NULL; }

Putting it all together {#directive_complete}
-----------------------

Now that we have our directives set up, and handlers configured for
them, we can assemble our module into one big file:

/\* mod\_example\_config\_simple.c: \*/ \#include \<stdio.h\> \#include
\"apr\_hash.h\" \#include \"ap\_config.h\" \#include \"ap\_provider.h\"
\#include \"httpd.h\" \#include \"http\_core.h\" \#include
\"http\_config.h\" \#include \"http\_log.h\" \#include
\"http\_protocol.h\" \#include \"http\_request.h\" /\*
==============================================================================
Our configuration prototype and declaration:
==============================================================================
\*/ typedef struct { int enabled; /\* Enable or disable our module \*/
const char \*path; /\* Some path to\...something \*/ int typeOfAction;
/\* 1 means action A, 2 means action B and so on \*/ } example\_config;
static example\_config config; /\*
==============================================================================
Our directive handlers:
==============================================================================
\*/ /\* Handler for the \"exampleEnabled\" directive \*/ const char
\*example\_set\_enabled(cmd\_parms \*cmd, void \*cfg, const char \*arg)
{ if(!strcasecmp(arg, \"on\")) config.enabled = 1; else config.enabled =
0; return NULL; } /\* Handler for the \"examplePath\" directive \*/
const char \*example\_set\_path(cmd\_parms \*cmd, void \*cfg, const char
\*arg) { config.path = arg; return NULL; } /\* Handler for the
\"exampleAction\" directive \*/ /\* Let\'s pretend this one takes one
argument (file or db), and a second (deny or allow), \*/ /\* and we
store it in a bit-wise manner. \*/ const char
\*example\_set\_action(cmd\_parms \*cmd, void \*cfg, const char \*arg1,
const char \*arg2) { if(!strcasecmp(arg1, \"file\")) config.typeOfAction
= 0x01; else config.typeOfAction = 0x02; if(!strcasecmp(arg2, \"deny\"))
config.typeOfAction += 0x10; else config.typeOfAction += 0x20; return
NULL; } /\*
==============================================================================
The directive structure for our name tag:
==============================================================================
\*/ static const command\_rec example\_directives\[\] = {
AP\_INIT\_TAKE1(\"exampleEnabled\", example\_set\_enabled, NULL,
RSRC\_CONF, \"Enable or disable mod\_example\"),
AP\_INIT\_TAKE1(\"examplePath\", example\_set\_path, NULL, RSRC\_CONF,
\"The path to whatever\"), AP\_INIT\_TAKE2(\"exampleAction\",
example\_set\_action, NULL, RSRC\_CONF, \"Special action value!\"), {
NULL } }; /\*
==============================================================================
Our module handler:
==============================================================================
\*/ static int example\_handler(request\_rec \*r) { if(!r-\>handler \|\|
strcmp(r-\>handler, \"example-handler\")) return(DECLINED);
ap\_set\_content\_type(r, \"text/plain\"); ap\_rprintf(r, \"Enabled:
%u\\n\", config.enabled); ap\_rprintf(r, \"Path: %s\\n\", config.path);
ap\_rprintf(r, \"TypeOfAction: %x\\n\", config.typeOfAction); return OK;
} /\*
==============================================================================
The hook registration function (also initializes the default config
values):
==============================================================================
\*/ static void register\_hooks(apr\_pool\_t \*pool) { config.enabled =
1; config.path = \"/foo/bar\"; config.typeOfAction = 3;
ap\_hook\_handler(example\_handler, NULL, NULL, APR\_HOOK\_LAST); } /\*
==============================================================================
Our module name tag:
==============================================================================
\*/ module AP\_MODULE\_DECLARE\_DATA example\_module = {
STANDARD20\_MODULE\_STUFF, NULL, /\* Per-directory configuration handler
\*/ NULL, /\* Merge handler for per-directory configurations \*/ NULL,
/\* Per-server configuration handler \*/ NULL, /\* Merge handler for
per-server configurations \*/ example\_directives, /\* Any directives we
may have for httpd \*/ register\_hooks /\* Our hook registering function
\*/ };

In our httpd.conf file, we can now change the hard-coded configuration
by adding a few lines:

ExampleEnabled On ExamplePath \"/usr/bin/foo\" ExampleAction file allow

And thus we apply the configuration, visit

/example

on our web site, and we see the configuration has adapted to what we
wrote in our configuration file.

Context aware configurations {#context}
============================

Introduction to context aware configurations {#context_intro}
--------------------------------------------

In Apache HTTP Server 2.4, different URLs, virtual hosts, directories
etc can have very different meanings to the user of the server, and thus
different contexts within which modules must operate. For example,
let\'s assume you have this configuration set up for mod\_rewrite:

\<Directory \"/var/www\"\> RewriteCond \"%{HTTP\_HOST}\"
\"\^example.com\$\" RewriteRule \"(.\*)\" \"http://www.example.com/\$1\"
\</Directory\> \<Directory \"/var/www/sub\"\> RewriteRule \"\^foobar\$\"
\"index.php?foobar=true\" \</Directory\>

In this example, you will have set up two different contexts for
mod\_rewrite:

Inside

/var/www

, all requests for

http://example.com

must go to

http://www.example.com

Inside

/var/www/sub

, all requests for

foobar

must go to

index.php?foobar=true

If mod\_rewrite (or the entire server for that matter) wasn\'t context
aware, then these rewrite rules would just apply to every and any
request made, regardless of where and how they were made, but since the
module can pull the context specific configuration straight from the
server, it does not need to know itself, which of the directives are
valid in this context, since the server takes care of this.

So how does a module get the specific configuration for the server,
directory or location in question? It does so by making one simple call:

example\_config \*config = (example\_config\*)

ap\_get\_module\_config

(r-\>per\_dir\_config, &example\_module);

That\'s it! Of course, a whole lot goes on behind the scenes, which we
will discuss in this chapter, starting with how the server came to know
what our configuration looks like, and how it came to be set up as it is
in the specific context.

Our basic configuration setup {#context_base}
-----------------------------

In this chapter, we will be working with a slightly modified version of
our previous context structure. We will set a

context

variable that we can use to track which context configuration is being
used by the server in various places:

typedef struct { char context\[256\]; char path\[256\]; int
typeOfAction; int enabled; } example\_config;

Our handler for requests will also be modified, yet still very simple:

static int example\_handler(request\_rec \*r) { if(!r-\>handler \|\|
strcmp(r-\>handler, \"example-handler\")) return(DECLINED);
example\_config \*config = (example\_config\*)
ap\_get\_module\_config(r-\>per\_dir\_config, &example\_module);
ap\_set\_content\_type(r, \"text/plain\"); ap\_rprintf(\"Enabled:
%u\\n\", config-\>enabled); ap\_rprintf(\"Path: %s\\n\", config-\>path);
ap\_rprintf(\"TypeOfAction: %x\\n\", config-\>typeOfAction);
ap\_rprintf(\"Context: %s\\n\", config-\>context); return OK; }

Choosing a context {#context_which}
------------------

Before we can start making our module context aware, we must first
define, which contexts we will accept. As we saw in the previous
chapter, defining a directive required five elements be set:

AP\_INIT\_TAKE1(\"exampleEnabled\", example\_set\_enabled, NULL,
RSRC\_CONF, \"Enable or disable mod\_example\"),

The

RSRC\_CONF

definition told the server that we would only allow this directive in a
global server context, but since we are now trying out a context aware
version of our module, we should set this to something more lenient,
namely the value

ACCESS\_CONF

, which lets us use the directive inside \<Directory\> and \<Location\>
blocks. For more control over the placement of your directives, you can
combine the following restrictions together to form a specific rule:

RSRC\_CONF

: Allow in .conf files (not .htaccess) outside \<Directory\> or
\<Location\>

ACCESS\_CONF

: Allow in .conf files (not .htaccess) inside \<Directory\> or
\<Location\>

OR\_OPTIONS

: Allow in .conf files and .htaccess when

AllowOverride Options

is set

OR\_FILEINFO

: Allow in .conf files and .htaccess when

AllowOverride FileInfo

is set

OR\_AUTHCFG

: Allow in .conf files and .htaccess when

AllowOverride AuthConfig

is set

OR\_INDEXES

: Allow in .conf files and .htaccess when

AllowOverride Indexes

is set

OR\_ALL

: Allow anywhere in .conf files and .htaccess

Using the server to allocate configuration slots {#context_pool}
------------------------------------------------

A much smarter way to manage your configurations is by letting the
server help you create them. To do so, we must first start off by
changing our

name tag

to let the server know, that it should assist us in creating and
managing our configurations. Since we have chosen the per-directory (or
per-location) context for our module configurations, we\'ll add a
per-directory creator and merger function reference in our tag:

module AP\_MODULE\_DECLARE\_DATA example\_module = {
STANDARD20\_MODULE\_STUFF, create\_dir\_conf, /\* Per-directory
configuration handler \*/ merge\_dir\_conf, /\* Merge handler for
per-directory configurations \*/ NULL, /\* Per-server configuration
handler \*/ NULL, /\* Merge handler for per-server configurations \*/
directives, /\* Any directives we may have for httpd \*/ register\_hooks
/\* Our hook registering function \*/ };

Creating new context configurations {#context_new}
-----------------------------------

Now that we have told the server to help us create and manage
configurations, our first step is to make a function for creating new,
blank configurations. We do so by creating the function we just
referenced in our name tag as the Per-directory configuration handler:

void \*create\_dir\_conf(apr\_pool\_t \*pool, char \*context) { context
= context ? context : \"(undefined context)\"; example\_config \*cfg =
apr\_pcalloc(pool, sizeof(example\_config)); if(cfg) { /\* Set some
default values \*/ strcpy(cfg-\>context, context); cfg-\>enabled = 0;
cfg-\>path = \"/foo/bar\"; cfg-\>typeOfAction = 0x11; } return cfg; }

Merging configurations {#context_merge}
----------------------

Our next step in creating a context aware configuration is merging
configurations. This part of the process particularly applies to
scenarios where you have a parent configuration and a child, such as the
following:

\<Directory \"/var/www\"\> ExampleEnabled On ExamplePath \"/foo/bar\"
ExampleAction file allow \</Directory\> \<Directory
\"/var/www/subdir\"\> ExampleAction file deny \</Directory\>

In this example, it is natural to assume that the directory

/var/www/subdir

should inherit the values set for the

/var/www

directory, as we did not specify an

ExampleEnabled

nor an

ExamplePath

for this directory. The server does not presume to know if this is true,
but cleverly does the following:

Creates a new configuration for

/var/www

Sets the configuration values according to the directives given for

/var/www

Creates a new configuration for

/var/www/subdir

Sets the configuration values according to the directives given for

/var/www/subdir

Proposes a merge

of the two configurations into a new configuration for

/var/www/subdir

This proposal is handled by the

merge\_dir\_conf

function we referenced in our name tag. The purpose of this function is
to assess the two configurations and decide how they are to be merged:

void \*merge\_dir\_conf(apr\_pool\_t \*pool, void \*BASE, void \*ADD) {
example\_config \*base = (example\_config \*) BASE ; /\* This is what
was set in the parent context \*/ example\_config \*add =
(example\_config \*) ADD ; /\* This is what is set in the new context
\*/ example\_config \*conf = (example\_config \*)
create\_dir\_conf(pool, \"Merged configuration\"); /\* This will be the
merged configuration \*/ /\* Merge configurations \*/ conf-\>enabled = (
add-\>enabled == 0 ) ? base-\>enabled : add-\>enabled ;
conf-\>typeOfAction = add-\>typeOfAction ? add-\>typeOfAction :
base-\>typeOfAction; strcpy(conf-\>path, strlen(add-\>path) ? add-\>path
: base-\>path); return conf ; }

Trying out our new context aware configurations {#context_example}
-----------------------------------------------

Now, let\'s try putting it all together to create a new module that is
context aware. First off, we\'ll create a configuration that lets us
test how the module works:

\<Location \"/a\"\> SetHandler example-handler ExampleEnabled on
ExamplePath \"/foo/bar\" ExampleAction file allow \</Location\>
\<Location \"/a/b\"\> ExampleAction file deny ExampleEnabled off
\</Location\> \<Location \"/a/b/c\"\> ExampleAction db deny ExamplePath
\"/foo/bar/baz\" ExampleEnabled on \</Location\>

Then we\'ll assemble our module code. Note, that since we are now using
our name tag as reference when fetching configurations in our handler, I
have added some prototypes to keep the compiler happy:

/\*\$6
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
\* mod\_example\_config.c
+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
\*/ \#include \<stdio.h\> \#include \"apr\_hash.h\" \#include
\"ap\_config.h\" \#include \"ap\_provider.h\" \#include \"httpd.h\"
\#include \"http\_core.h\" \#include \"http\_config.h\" \#include
\"http\_log.h\" \#include \"http\_protocol.h\" \#include
\"http\_request.h\" /\*\$1
\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~
Configuration structure
\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~
\*/ typedef struct { char context\[256\]; char path\[256\]; int
typeOfAction; int enabled; } example\_config; /\*\$1
\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~
Prototypes
\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~
\*/ static int example\_handler(request\_rec \*r); const char
\*example\_set\_enabled(cmd\_parms \*cmd, void \*cfg, const char \*arg);
const char \*example\_set\_path(cmd\_parms \*cmd, void \*cfg, const char
\*arg); const char \*example\_set\_action(cmd\_parms \*cmd, void \*cfg,
const char \*arg1, const char \*arg2); void
\*create\_dir\_conf(apr\_pool\_t \*pool, char \*context); void
\*merge\_dir\_conf(apr\_pool\_t \*pool, void \*BASE, void \*ADD); static
void register\_hooks(apr\_pool\_t \*pool); /\*\$1
\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~
Configuration directives
\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~
\*/ static const command\_rec directives\[\] = {
AP\_INIT\_TAKE1(\"exampleEnabled\", example\_set\_enabled, NULL,
ACCESS\_CONF, \"Enable or disable mod\_example\"),
AP\_INIT\_TAKE1(\"examplePath\", example\_set\_path, NULL, ACCESS\_CONF,
\"The path to whatever\"), AP\_INIT\_TAKE2(\"exampleAction\",
example\_set\_action, NULL, ACCESS\_CONF, \"Special action value!\"), {
NULL } }; /\*\$1
\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~
Our name tag
\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~
\*/ module AP\_MODULE\_DECLARE\_DATA example\_module = {
STANDARD20\_MODULE\_STUFF, create\_dir\_conf, /\* Per-directory
configuration handler \*/ merge\_dir\_conf, /\* Merge handler for
per-directory configurations \*/ NULL, /\* Per-server configuration
handler \*/ NULL, /\* Merge handler for per-server configurations \*/
directives, /\* Any directives we may have for httpd \*/ register\_hooks
/\* Our hook registering function \*/ }; /\*
=======================================================================================================================
Hook registration function
=======================================================================================================================
\*/ static void register\_hooks(apr\_pool\_t \*pool) {
ap\_hook\_handler(example\_handler, NULL, NULL, APR\_HOOK\_LAST); } /\*
=======================================================================================================================
Our example web service handler
=======================================================================================================================
\*/ static int example\_handler(request\_rec \*r) { if(!r-\>handler \|\|
strcmp(r-\>handler, \"example-handler\")) return(DECLINED);
/\*\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\*/
example\_config \*config = (example\_config \*)
ap\_get\_module\_config(r-\>per\_dir\_config, &example\_module);
/\*\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\*/
ap\_set\_content\_type(r, \"text/plain\"); ap\_rprintf(r, \"Enabled:
%u\\n\", config-\>enabled); ap\_rprintf(r, \"Path: %s\\n\",
config-\>path); ap\_rprintf(r, \"TypeOfAction: %x\\n\",
config-\>typeOfAction); ap\_rprintf(r, \"Context: %s\\n\",
config-\>context); return OK; } /\*
=======================================================================================================================
Handler for the \"exampleEnabled\" directive
=======================================================================================================================
\*/ const char \*example\_set\_enabled(cmd\_parms \*cmd, void \*cfg,
const char \*arg) {
/\*\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\*/
example\_config \*conf = (example\_config \*) cfg;
/\*\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\*/
if(conf) { if(!strcasecmp(arg, \"on\")) conf-\>enabled = 1; else
conf-\>enabled = 0; } return NULL; } /\*
=======================================================================================================================
Handler for the \"examplePath\" directive
=======================================================================================================================
\*/ const char \*example\_set\_path(cmd\_parms \*cmd, void \*cfg, const
char \*arg) {
/\*\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\*/
example\_config \*conf = (example\_config \*) cfg;
/\*\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\*/
if(conf) { strcpy(conf-\>path, arg); } return NULL; } /\*
=======================================================================================================================
Handler for the \"exampleAction\" directive ; Let\'s pretend this one
takes one argument (file or db), and a second (deny or allow), ; and we
store it in a bit-wise manner.
=======================================================================================================================
\*/ const char \*example\_set\_action(cmd\_parms \*cmd, void \*cfg,
const char \*arg1, const char \*arg2) {
/\*\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\*/
example\_config \*conf = (example\_config \*) cfg;
/\*\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\*/
if(conf) { { if(!strcasecmp(arg1, \"file\")) conf-\>typeOfAction = 0x01;
else conf-\>typeOfAction = 0x02; if(!strcasecmp(arg2, \"deny\"))
conf-\>typeOfAction += 0x10; else conf-\>typeOfAction += 0x20; } }
return NULL; } /\*
=======================================================================================================================
Function for creating new configurations for per-directory contexts
=======================================================================================================================
\*/ void \*create\_dir\_conf(apr\_pool\_t \*pool, char \*context) {
context = context ? context : \"Newly created configuration\";
/\*\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\*/
example\_config \*cfg = apr\_pcalloc(pool, sizeof(example\_config));
/\*\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\*/
if(cfg) { { /\* Set some default values \*/ strcpy(cfg-\>context,
context); cfg-\>enabled = 0; memset(cfg-\>path, 0, 256);
cfg-\>typeOfAction = 0x00; } } return cfg; } /\*
=======================================================================================================================
Merging function for configurations
=======================================================================================================================
\*/ void \*merge\_dir\_conf(apr\_pool\_t \*pool, void \*BASE, void
\*ADD) {
/\*\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\*/
example\_config \*base = (example\_config \*) BASE; example\_config
\*add = (example\_config \*) ADD; example\_config \*conf =
(example\_config \*) create\_dir\_conf(pool, \"Merged configuration\");
/\*\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\*/
conf-\>enabled = (add-\>enabled == 0) ? base-\>enabled : add-\>enabled;
conf-\>typeOfAction = add-\>typeOfAction ? add-\>typeOfAction :
base-\>typeOfAction; strcpy(conf-\>path, strlen(add-\>path) ? add-\>path
: base-\>path); return conf; }

Summing up {#summary}
==========

We have now looked at how to create simple modules for Apache HTTP
Server 2.4 and configuring them. What you do next is entirely up to you,
but it is my hope that something valuable has come out of reading this
documentation. If you have questions on how to further develop modules,
you are welcome to join our

mailing lists

or check out the rest of our documentation for further tips.

Some useful snippets of code {#snippets}
============================

Retrieve variables from POST form data {#get_post}
--------------------------------------

typedef struct { const char \*key; const char \*value; } keyValuePair;
keyValuePair \*readPost(request\_rec \*r) { apr\_array\_header\_t
\*pairs = NULL; apr\_off\_t len; apr\_size\_t size; int res; int i = 0;
char \*buffer; keyValuePair \*kvp; res = ap\_parse\_form\_data(r, NULL,
&pairs, -1, HUGE\_STRING\_LEN); if (res != OK \|\| !pairs) return NULL;
/\* Return NULL if we failed or if there are is no POST data \*/ kvp =
apr\_pcalloc(r-\>pool, sizeof(keyValuePair) \* (pairs-\>nelts + 1));
while (pairs && !apr\_is\_empty\_array(pairs)) { ap\_form\_pair\_t
\*pair = (ap\_form\_pair\_t \*) apr\_array\_pop(pairs);
apr\_brigade\_length(pair-\>value, 1, &len); size = (apr\_size\_t) len;
buffer = apr\_palloc(r-\>pool, size + 1);
apr\_brigade\_flatten(pair-\>value, buffer, &size); buffer\[len\] = 0;
kvp\[i\].key = apr\_pstrdup(r-\>pool, pair-\>name); kvp\[i\].value =
buffer; i++; } return kvp; } static int example\_handler(request\_rec
\*r) { /\*\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\*/ keyValuePair
\*formData; /\*\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\*/ formData
= readPost(r); if (formData) { int i; for (i = 0; &formData\[i\]; i++) {
if (formData\[i\].key && formData\[i\].value) { ap\_rprintf(r, \"%s =
%s\\n\", formData\[i\].key, formData\[i\].value); } else if
(formData\[i\].key) { ap\_rprintf(r, \"%s\\n\", formData\[i\].key); }
else if (formData\[i\].value) { ap\_rprintf(r, \"= %s\\n\",
formData\[i\].value); } else { break; } } } return OK; }

Printing out every HTTP header received {#headers_out}
---------------------------------------

static int example\_handler(request\_rec \*r) {
/\*\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\*/
const apr\_array\_header\_t \*fields; int i; apr\_table\_entry\_t \*e =
0;
/\*\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\*/
fields = apr\_table\_elts(r-\>headers\_in); e = (apr\_table\_entry\_t
\*) fields-\>elts; for(i = 0; i \< fields-\>nelts; i++) { ap\_rprintf(r,
\"%s: %s\\n\", e\[i\].key, e\[i\].val); } return OK; }

Reading the request body into memory {#request_body}
------------------------------------

static int util\_read(request\_rec \*r, const char \*\*rbuf, apr\_off\_t
\*size) { /\*\~\~\~\~\~\~\~\~\*/ int rc = OK; /\*\~\~\~\~\~\~\~\~\*/
if((rc = ap\_setup\_client\_block(r, REQUEST\_CHUNKED\_ERROR))) {
return(rc); } if(ap\_should\_client\_block(r)) {
/\*\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\*/
char argsbuffer\[HUGE\_STRING\_LEN\]; apr\_off\_t rsize, len\_read, rpos
= 0; apr\_off\_t length = r-\>remaining;
/\*\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\*/
\*rbuf = (const char \*) apr\_pcalloc(r-\>pool, (apr\_size\_t) (length +
1)); \*size = length; while((len\_read = ap\_get\_client\_block(r,
argsbuffer, sizeof(argsbuffer))) \> 0) { if((rpos + len\_read) \>
length) { rsize = length - rpos; } else { rsize = len\_read; }
memcpy((char \*) \*rbuf + rpos, argsbuffer, (size\_t) rsize); rpos +=
rsize; } } return(rc); } static int example\_handler(request\_rec \*r) {
/\*\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\*/ apr\_off\_t size; const char
\*buffer; /\*\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\~\*/ if(util\_read(r,
&buffer, &size) == OK) { ap\_rprintf(r, \"We read a request body that
was %\" APR\_OFF\_T\_FMT \" bytes long\", size); } return OK; }
